# plot_seq_tools.py

from auxiliary import cm2inch
import os, numpy as np, pandas as pd, sys, re, math
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.ticker import MultipleLocator
from matplotlib.ticker import MaxNLocator
from extract_pcr import getwavelength
from find_tools import findfolders_abspath, findfiles_abspath



class Base():
  def __init__(self, nb, angle_range = None, minute_range = None, printsize = "two_in_docx"):
    self.df = nb.datdf
    self.outpath = nb.outpath
    self.system = nb.exp_name
    self.printsize = printsize
    self.path = os.getcwd()
    self.minute_range = minute_range
    """ checking if dataframe is empty """
    if self.df.empty:
      print("--no data frames have been found!")
      return -1

    self.xvalues = self.df.index.values
    if type(angle_range) == type(tuple()):
      start,end = angle_range
      mymask1 = self.xvalues > start
      mymask2 = self.xvalues < end
      mymask = mymask1 == mymask2
      self.df = self.df.iloc[mymask]
    self.xvalues = self.df.index.values


    self.defining_sizes()


  def celcius2latex(self, mystring = ""):
    """ returns a latexified string to be used in plt.
    e.g. 500oC --> 500$^o$C """
    return re.sub("(\d)oC","\g<1>°C",mystring)

  def setting_qspace(self):
    for self.myfolder in findfolders_abspath():
      pcrs = findfiles_abspath(".pcr", path = self.myfolder)
      if len(pcrs) > 0:
        self.wavelength = getwavelength(pcrs[0])
        break
    vfunc = np.vectorize(math.sin)
    self.df.index = math.pi * 4 / self.wavelength * ( vfunc( math.pi / 180 / 2 * self.df.index.values) )
    os.chdir(self.path)
    """ this is mask for Q-space

    EDIT: use xlim instead! """
    self.xlow = 2.8
    self.xhigh = 3.2
    mymask1 = self.df.index.values > self.xlow
    mymask2 = self.df.index.values < self.xhigh
    mymask = mymask1 == mymask2
    self.df = self.df.iloc[mymask]
    self.xvalues = self.df.index.values


  def meshgrid(self):
    """prepare meshgrids and plotting parameters"""
    self.angle_mesh,self.time_mesh = np.meshgrid(self.xvalues, self.yvalues)
    self.signals = self.df.values.T


  def plot1(self):
    """ doing the actual plotting """
    fig = plt.figure(figsize=self.figsize)

    """higher number for nbin yields higher level of detail in plot"""
    self.nbins = [100] # 10 does not show the weak features... no difference between 100 and 200
    nbin = self.nbins[0]
    if len(self.signals) == 0: # the specified range for plotting includes no values.
      return -1
    levels = MaxNLocator(nbins=nbin).tick_values(self.signals.min(), self.signals.max())
    style = self.styles[0]
    plt_cmap = plt.get_cmap(style)
    CS = plt.contourf(self.angle_mesh, self.time_mesh, self.signals,
                      cmap=plt_cmap,
                      levels = levels
                      )

    """ getting the axis object """
    self.baseax = plt.gca()

    """ adding colorbar """
    matplotlib.rcParams.update({'font.size': self.textsize})
    self.mybar = plt.colorbar(ticks = [])
    # self._mybarvalues = self.mybar._tick_data_values
    # self.mybar._tick_data_values = [1000000 for x in self._mybarvalues]

    """ setting up xlabel and ylabel """
    plt.xlabel(r'Q-space [$\AA^{-1}$]', size = self.xlabelsize)
    ylabel = plt.ylabel('time [min]', size = self.ylabelsize)

    """ changing the size of x ticks"""
    for tick in self.baseax.xaxis.get_major_ticks():
      tick.label.set_fontsize(self.xticksize)

    for tick in self.baseax.yaxis.get_major_ticks():
      tick.label.set_fontsize(self.yticksize)

    """ set padding of first y-axis minute ticks and change length """
    self.baseax.get_yaxis().set_tick_params(pad=1, length = 2)

    """ set padding of x-axis angle ticks and change length"""
    self.baseax.get_xaxis().set_tick_params(pad=1, length = 2)


  def rotating_ylabels(self, printsize = "two_in_docx"):
    """ rotating and repositioning the ylabels """
    if self.printsize == "thirdpage":
      self.baseax.yaxis.set_label_coords(0,1.05)
      ylabel.set_rotation(0)




  def defining_sizes(self):
    """ pypot uses inches... """
    inch2cm = 1.0 / 2.54
    """ horizontal length of graph """
    dimension = {"onepage": 17 * inch2cm, "two_in_docx": 8 * inch2cm, "thirdpage": 5.0 * inch2cm, "four_in_docx": 4.2 * inch2cm}
    choosen_dimension = dimension[self.printsize]

    """ sizes of figure, ticks, labels and legends """
    height_factor = 1.3
    if self.printsize == "two_in_docx":
      self.figsize = (choosen_dimension,choosen_dimension * height_factor)
      self.textsize = 8
      self.titlesize = self.textsize + 1
      self.xlabelsize = self.textsize
      self.ylabelsize = self.textsize
      self.xticksize = self.textsize
      self.yticksize = self.textsize -1
      self.linewidth = 0.8
      self.ylabelpos_pre = (0.5, 0.5)
      self.ylabelpos = (-0.12, 0.5)
      self.ytick_placeholder = "1000"
    elif self.printsize == "thirdpage":
      self.figsize = (choosen_dimension,choosen_dimension * height_factor)
      self.textsize = 6
      self.titlesize = self.textsize
      self.xlabelsize = self.textsize
      self.ylabelsize = self.textsize
      self.xticksize = self.textsize
      self.yticksize = self.textsize
    elif self.printsize == "four_in_docx":
      self.figsize = (choosen_dimension,choosen_dimension * height_factor)
      self.textsize = 5
      self.titlesize = self.textsize + 1
      self.xlabelsize = self.textsize
      self.ylabelsize = self.textsize
      self.xticksize = self.textsize
      self.yticksize = self.textsize
      self.ylabelpos_pre = (0.5, 0.5)
      self.ylabelpos = (-0.2, 0.5)
      self.linewidth = 0.4
      self.ytick_placeholder = "10000"

  def title(self):
    """ setting title """
    # title_ypos = 1.0 # percent of canvas.
    plt.title( self.celcius2latex(self.system), size = self.titlesize)#, y = title_ypos)

  def tight_layout(self):
    self.myyticks = self.baseax.get_yticks()
    plt.yticks(self.myyticks, [self.ytick_placeholder for ytick in self.myyticks])
    self.baseax.yaxis.set_label_coords(*self.ylabelpos_pre)
    """ enable tight layout """
    plt.tight_layout()  # do not use with altered ylabels.
    replacement = [int(y) for y in self.myyticks]
    replacement[replacement.index(0)] = ""
    plt.yticks(self.myyticks, replacement)
    plt.ylim(min(self.yvalues),max(self.yvalues))
    self.myxticks = self.baseax.get_xticks()
    plt.xticks(np.arange(self.xlow,self.xhigh,0.2))
    """ setting xlim """
    plt.xlim(self.xlow, self.xhigh)
    self.baseax.yaxis.set_label_coords(*self.ylabelpos)
    """ resetting colorbar tick values """
    # self.mybar._tick_data_values = self._mybarvalues

    plt.tick_params(
    axis='y',          # changes apply to the x-axis
    which='both',      # both major and minor ticks are affected
    left='off',      # ticks along the bottom edge are off
    right='off',         # ticks along the top edge are off
    # labelbottom='off'
    ) # labels along the bottom edge are off

  def savefig(self):
    """ saving plot , outfolder has already been created """
    figname = self.system + ".png"
    plt.savefig( os.path.join(self.outpath,figname), dpi = 250 )
    plt.close() # save pc from crashing!




class Dioptas(Base):
  def __init__(self, nb, angle_range = None, minute_range = None, printsize = "two_in_docx"):
    args = (nb, angle_range, minute_range, printsize)
    super(Dioptas, self).__init__(*args)


    index = "angle"
    """ index was already set... """
    # df = df.set_index(index)
    """ prepare minutes array"""
    duration_frame = 5 # seconds, blindly hardcoded...
    nb_of_frames = len(self.df.columns.values)

    """ converts the frames, e.g. frame18, frame19, .. frameN to integer list [18, 19, .., N] """
    myframes = [int(index.strip("frame")) for index in self.df.columns]

    """ blindly defines the first frame e.g. frame193 to be zero point in time """
    myframes = [integ - myframes[0] for integ in myframes]
    self.yvalues = [frame * (duration_frame / 60.0) for frame in myframes]
    self.yvalues = np.array(self.yvalues)

    """ truncate DataFrame if user asks for it
    truncation is done by indicating what (x1,x2) minutes to include"""
    if type(self.minute_range) is type(tuple()):
      """ this is to cut the time, NOT the angles!! """
      lower_bound, higher_bound = self.minute_range
      con1 = self.yvalues > lower_bound
      self.df = self.df.iloc[:,con1]
      self.yvalues = self.yvalues[con1]
      con2 = self.yvalues < higher_bound
      self.df = self.df.iloc[:,con2]
      self.yvalues = self.yvalues[con2]

      """check if any values are left"""
      if len(self.yvalues) == 0:
        print("no values to plot for",self.minute_range, "in", path)
        return 1


    """ these are styles that have better contrast:
    choosen_cmaps = "gnuplot gnuplot2 jet ocean viridis Paired Oranges PuBu " +\
        "PuBuGn PuRd Reds YlGn YlGrBu YlOrRd cool"
    """
    self.styles = [style for style in 'jet'.split()]

  def setting_minutes(self):
    """ changing yticks and their size"""
    self.minutes_simplified = [int(i) for i in self.yvalues]
    plt.yticks(self.minutes_simplified)


  def tick_spacing(self):
    """ showing only time-yticklabel every tick_spacing """
    limit = max(self.yvalues) - min(self.yvalues)
    if limit <= 20:
      tick_spacing = 1 # min
    if limit <= 60:
      tick_spacing = 5 # min
    else:
      tick_spacing = 10 # min

    self.baseax.yaxis.set_major_locator(MultipleLocator(tick_spacing))
    # """ this is important when the 2theta view is truncated... """
    # self.baseax.xaxis.set_major_locator(MultipleLocator(0.5))

  def run(self):

    self.setting_qspace()

    self.meshgrid()

    if self.plot1() == -1:
      return -1

    self.setting_minutes()

    self.tick_spacing()

    self.rotating_ylabels()

    self.title()

    self.tight_layout()

    self.savefig()

    return 0







class Dmc(Base):
  def __init__(self, nb, angle_range = None, minute_range = None, printsize = "two_in_docx"):
    args = (nb, angle_range, minute_range, printsize)
    super(Dmc, self).__init__(*args)

    """ limiting dataframe to first x minutes """
    if type(minute_range) == type(tuple()):
      xmin,xmax = minute_range
      self.df = self.df.loc[:, self.df.columns <= xmax]
      self.df = self.df.loc[:, self.df.columns >= xmin]
    if type(minute_range) == type(int):
      index_to_last = minute_range
      self.df = self.df.iloc[:,minute_range:]

    """ changing yticks and their size"""
    self.yvalues = self.df.columns.values
    self.minutes_simplified = [int(float(i)) for i in self.yvalues]

    self.styles = [style for style in 'gnuplot'.split()]
    """ these are styles with better contrast:
    choosen_cmaps = "gnuplot gnuplot2 jet ocean viridis Paired Oranges PuBu " +\
        "PuBuGn PuRd Reds YlGn YlGrBu YlOrRd cool"
    """

  def setting_minutes(self):
    """ alternating yticks, skipping every other ytick """
    if 1:
      # temp_simplified = [int(t) for i,t in enumerate(temperatures) if i%2==0]
      self.minutes_simplified = [int(t) for i,t in enumerate(self.minutes_simplified) if i%2==0]
      new_yvalues = [t for i,t in enumerate(self.yvalues) if i%2==0]
    else:
      new_yvalues = self.yvalues

    plt.yticks(new_yvalues, self.minutes_simplified)


  def remove_sapphire_peak(self):
    """ this is to remove the single crystal sapphire peak... """
    ab = self.df.index < 61.6
    ba = self.df.index > 62.5
    mymask = ab == ba
    """ pandas gives warning about possible bad use
    of setting values of a slice-view. I don't know how to silence it... """
    self.df[mymask] = 0
    """ this command gives no warnings. but it might be deprecented in future version?
    https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.set_value.html
    the below is being deprecated. I will continue using the above with warning.
    """
    # df.set_value(mymask,df.columns,0)

  def dotted_lines(self):
    """ adding visual clue for each frame """
    for y in self.yvalues: # dotted lines
      plt.plot([min(self.xvalues), max(self.xvalues)], [y, y], "--r", linewidth = self.linewidth, alpha = 0.6) # dotted lines

  def run(self):

    """ unique for Dmc """
    self.remove_sapphire_peak()

    self.setting_qspace()

    self.meshgrid()

    if self.plot1() == -1:
      return -1

    self.setting_minutes()

    """ unique for Dmc """
    self.dotted_lines()

    self.rotating_ylabels()

    self.title()

    self.tight_layout()

    self.savefig()

    return 0














# end



import os, sys,re, pandas as pd, matplotlib.pyplot as plt
sys.path.append(r"C:\AU-PHD\General_Data\_python\PyPi\magnetmatter\magnetmatter\modules")

class Container():
  """ note, if round brackets are _not_ used, no new instance
  of the class is created.
  _Without_ round brackets the first class object is used over and over.
  my code worked for generating the seq.csv file, but that was
  a pure coincidence. """

  def __init__(self, value = None, error = None):
    """ construct with None was necessary to avoid python to
    reuse an earlier initilized Container object.
    The expressions in the initializing part of a function is
    only called _once_ over an entire process. """
    self.value = value if value != None else []
    self.error = error if error != None else []

  def iterators(self):
    """ returns a tuple of iterators for the value and error lists """
    return (iter(self.value), iter(self.error))

  def __repr__(self):
    """ determines what is shown with built-in repr() or using interactive prompt """
    return "size of container (value, error) = (" + str(len(self.value)) + "," + str(len(self.value)) + ")"

  def __str__(self):
    """ determines what is shown with built-in print() """
    return repr(self)

  def __getitem__(self,key):
    """  overwrites default getitem command, e.g.
    list slicing alist[0], alist[0:10], alist[::-1] etc. """
    return Container(self.value[key], self.error[key])

  """ legacy code... """
  # def __lt__(self, object):
  #   """ define the lesser than < operator  """
  #   return self.value[0] < object.value[0]
  #
  # def __gt__(self,object):
  #   """ define the greater than > operator  """
  #   return self.value[0] > object.value[0]
  #
  # def __eq__(self, object):
  #   """ define the == operator """
  #   return self.value == object.value

class Mydict():
  """  """
  def __init__(self):
    """  """
    self.mydict = {}

  def check_repeating_elements(self):
    """ checking if multiple runs of sequential refinement have been
    carried out (refinement results are stupidly appended to existing .seq file) """

    mylist = self.mydict["NEW_REFINEMENT"].value
    # print("mylist =",mylist)
    # import pdb; pdb.set_trace();
    first_element = mylist[0]
    first_index = 0
    for num, element in enumerate(mylist):
      """ if sequential refinement has been carried out a second or more times
      the .seq file is just appended by FullProf, therefore introducion
      multiple occurences of same frame number. """
      if first_element == element:
        first_index = num
    # import pdb; pdb.set_trace();
    for key,item in self.mydict.items():
      # print(self.mydict[key].value[0:10])
      self.mydict[key] = self.mydict[key][first_index:]
      # print(self.mydict[key].value[0:10])
      # print("key =",key)


  def check_reverse(self):
    """ checking if sequential refinement was carried out
    from last to first frame.
    if so, we reverse the sequence of all parameters """
    mylist = self.mydict["NEW_REFINEMENT"].value
    first_element = mylist[0]
    first_index = 0
    if int(mylist[0]) > int(mylist[-1]):
      """ reversing the list of every refined parameter. """
      for key,item in self.mydict.items():
        self.mydict[key] = self.mydict[key][::-1]


  def add(self, var, value, error):
    """ adds <value, error> to the self.mydict under the tag <var> """
    if var in self.mydict.keys():
      self.mydict[var].value.append(value)
      self.mydict[var].error.append(error)
    else:
      self.mydict[var] = Container()
      self.add(var,value,error)


  def dump(self, workingdir = ""):
    """ generates a csv readable by pandas

    ignores refined background parameters
    """
    workingdir = os.getcwd() if workingdir == "" else workingdir
    csvpath = os.path.join(workingdir, "_seq.csv")
    handle = open(csvpath,"w")
    myiterators = []
    writeval = ""
    """ checking if multiple sequential refinements has occured """
    self.check_repeating_elements()
    """ checking (and correcting) if lists should be reversed
    (i.e. sequential refinement was done from last to first frame) """
    self.check_reverse()
    """ writing refined parameters to .csv file """
    for key in self.mydict.keys():
      """ first writing all column names """
      if key.startswith("Bck"): continue
      writeval += key + ","
      val_ite, err_ite = self.mydict[key].iterators()
      myiterators.append(val_ite)
      if not key in ["NEW_REFINEMENT", "GLOBAL_CHI_2"]:
        writeval += key + "_err,"
        myiterators.append(err_ite)
    writeval = writeval[:-1] + "\n"
    handle.write(writeval)
    writeval = ""
    # import pdb; pdb.set_trace();
    # import pdb; pdb.set_trace();
    while True:
      """ write all values and errors """
      try:
        for iterator in myiterators:
          writeval += str( iterator.__next__() ) + ","
        writeval = writeval[:-1] + "\n"
        if "NaN" in writeval:
          continue
        handle.write(writeval)
      except StopIteration:
        break
      finally:
        writeval = ""

    """ managing dangling file """
    handle.close()

    return csvpath


  def __repr__(self):
    retval = "keys of Mydict (" + str( len(self.mydict) ) + ") = ["
    for key in self.mydict.keys():
      retval += key + ",\n"
    retval += "\b\b] "
    return retval

  def __str__(self):
    return repr(self)



def seq2csv(csvpath):
  """ what is iterated over? Example:

  NEW_REFINEMENT       frame616
  NUMPAR    13
  N_PATT     1
  NPHASE     2
  PARVAL   408      0.0000        0
  GLOBAL_CHI_2      0.561820
       1         R_Bragg(%):     2.3198311         2.5157762
       1   Unit_Cell_Volume:     23.899607        0.95912709E-03
       1    Weight_Fraction:     13.881724        0.16681060
       2         R_Bragg(%):     8.1540928         5.6447506
       2   Unit_Cell_Volume:     599.82324        0.22575069E-01
       2    Weight_Fraction:     86.118271        0.61252558
       1     Cell_A_ph1_pat1     2.8804715        0.66740329E-04
       2     Cell_A_ph2_pat1     8.4334984        0.18325378E-03
       3           Zero_pat1   -0.35518815E-02    0.32384731E-03
       4          Bck_0_pat1     45.707390        0.18409243
       5          Bck_1_pat1    -31.257633        0.30232862
       6          Bck_2_pat1     13.554113        0.27447793
       7          Bck_3_pat1    -8.8670788        0.26058340
       8          Bck_4_pat1     3.7743287        0.22627439
       9          Bck_5_pat1    -1.3671141        0.21322817
      10      Scale_ph1_pat1    0.70008887E-05    0.77171720E-07
      11      Y-cos_ph1_pat1    0.82928445E-02    0.48348296E-03
      12      Scale_ph2_pat1    0.12760545E-06    0.67161066E-09
      13      Y-cos_ph2_pat1    0.30908650E-01    0.36040161E-03

  """
  # os.chdir(r"C:\AU-PHD\General_Data\Hamburg_2016May\gammaFe2O3_integrated\done_refined")
  workingdir = os.path.dirname(csvpath)
  mydict = Mydict()
  with open(csvpath,"r") as reading:
    for line in reading:
      found = re.match("\w",line)
      if found:
        found = re.match("GLOBAL_CHI_2\s+([^\s]+)",line)
        if found:
          found = found.group(1)
          mydict.add("GLOBAL_CHI_2", found, 0)
        found = re.match("NEW_REFINEMENT\s+[^\s]+?(\d+)$",line)
        if found:
          found = found.group(1)
          # print("found =",found)
          # import pdb; pdb.set_trace();
          mydict.add("NEW_REFINEMENT", found, 0)
      else:
        found = re.match("\s+(\d+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*", line)
        if found:
          number = found.group(1)
          var = found.group(2)
          var = "".join( re.split("_pat\d+",var) )
          value = found.group(3)
          error = found.group(4)
          if var.endswith(":"):
            var = var[:-1] + "_ph" + str(number)
          var = re.sub("\(%\)","", var)
          mydict.add(var, value, error)
  csvname = mydict.dump(workingdir)
  return csvname

def visualize_csv(csvpath):
  """ plots every refined parameter from FullProf that is found in csv

  csvpath:
    path to comma-separated-file
  """
  # plt.figure()
  df = pd.read_csv(csvpath, index_col = "NEW_REFINEMENT")
  for colname in df.columns:
    try:
      if colname.endswith("_err"): continue
      if colname + "_err" in df.columns:
        df.plot(y=colname, yerr = colname + "_err")
      else:
        df.plot(y=colname)
      plt.savefig(colname + ".png")
      plt.close()
    except Exception as e:
      print(e)
      import pdb; pdb.set_trace();
  return



import glob, os
import moviepy.editor as mpy

class Video():
  def __init__(path = ""):
    oldpath = os.getcwd()
    path = os.getcwd() if path == "" else path
    os.chdir(path)
    fps_read_images = 1
    fps_video = 10 # not sure if this does any difference...
    file_list = glob.glob('*.png') # Get all the pngs in the current directory
    list.sort(file_list, key=lambda x: x.split('to')[0])
    clip = mpy.ImageSequenceClip(file_list, fps=fps_read_images)

    # videoclip = mpy.VideoFileClip("some_animation.gif")
    clip.write_videofile(os.path.join(path,"..",'video.mp4'), fps=fps_video)
    # clip.write_gif('gif.gif', fps=fps_gif)
    os.chdir(oldpath)



# plot_seq.py

import shutil, sys, os, pandas as pd, re

""" magnetmatter modules """
from find_tools import findfolders_abspath, findfiles_abspath, findfile_abspath
from find_tools import findfolders, findfiles, findfile
import plot_predata
import subplots_seq_tools
import seqfile2seqcsv
import datfiles2datcsv
import extract_pcr
import seqcsv2sizecsv
import pdb; trace = pdb.set_trace




def plot_seq(workingdir = ""):
  """ workingdir is ./root of project.
  Searches all subfolders,
  creates plots and store in ./root/_mytest in separate experiment folders.

  this function executes:
    clean_output_folder
    plot_predata
    plot_refined_data
    plot_piecewise_refinements
    clean_after_plotting
  """

  class Notebook():
    """ reduce the need to pass several arguments to different functions """
    def __init__(self):
      self.lorsizelist = []
      # self.newlist = [] # seen used in plot_seq_tools for seqdf.
      self.phases = []
      self.wavelength = 0
      self.seqdf = pd.DataFrame
      self.sizedf = pd.DataFrame
      self.datdf = pd.DataFrame
      self.outpath = "folder where all graphs are created in subfolders"
      self.suboutpath = "path to subfolder in outpath"
      self.workingdir = "parent folder of all experiments"
      self.experiments = []
      self.exp_path = "current experiment"
      self.exp_name = "current experiment"
      self.ref_path = "current sequential refinement"
      self.ref_name = "plain name of the folder of sequential refinement"
      self.datdftype = "current types include dmc and dioptas"
      # self.plt_border_settings = "these are messed up when doing continues plotting combining contourf and plt.style.use"
      self.refinement_folders = []
      self.TwoByTwo = "subplotmanager for 2 x 2 subplot"
      self.TwoByOne = "subplotmanager for 1 x 2 subplots"
      self.TwoByTwoTruncated = "not really sure what this one does..."




  nb = Notebook()
  nb.workingdir = workingdir if workingdir != "" else os.getcwd()
  nb.outpath = os.path.join(nb.workingdir, "_mytest")

  """ important: setting working directory """
  os.chdir(nb.workingdir)

  """ cleaning up output folder """
  retval = clean_output_folder(nb)

  """ looping over all folders with presumed experimental data
  if such data is absent or something goes wrong,
  we skip the specific folder """
  exp_folders_to_skip = []
  for nb.exp_path in findfolders_abspath():
    nb.exp_name = os.path.basename(nb.exp_path)
    os.chdir(nb.exp_path)

    # if nb.exp_name == "FS_g18nm_P021_400oC_b":
    #   print("nb.exp_name =",nb.exp_name)
    #   import pdb; pdb.set_trace();
    """ ignore certain experiments. should be removed later """
    if os.path.basename(nb.exp_path) in [exp for exp in "400oC_a 400oC_c FS_g18nm_P021_500oC_a plt_styles archieve".split()]:
      print("    skipping current nb.exp_name = ", nb.exp_name)
      exp_folders_to_skip.append(-1)
      continue

    """ important: setting working folder """
    os.chdir(nb.exp_path)

    """ defining the folder where plots will be saved """
    nb.suboutpath = os.path.join(nb.workingdir, nb.outpath, os.path.basename(nb.exp_path))
    if not os.path.exists(nb.suboutpath):
      os.makedirs(nb.suboutpath)

    """ reading or generate datcsv """
    retval = read_predata(nb)
    if retval == -1:
      print("read_predata went wrong in nb.exp_name = ", nb.exp_name)
      exp_folders_to_skip.append(retval)
      continue

    """ trying to plot predata.
    We expect dioptas or DMC .dat files. """
    retval = plot_predata_handler(nb)
    """ something went wrong in plot_predata,
    we skip this experimental folder altogehter """
    exp_folders_to_skip.append(retval)
    """ note that plotting predata and refined data
    are separated. apparently the plt.style.use interfered
    with the formatting of all following created figures.
    therefore we have these two separate for-loops. """

  nb.TwoByTwo          = subplots_seq_tools.SubplotManager(nb, nrows=2,ncols=2)
  nb.TwoByOne          = subplots_seq_tools.SubplotManager(nb, nrows=2,ncols=1)
  nb.TwoByTwoTruncated = subplots_seq_tools.SubplotManager(nb, nrows=2,ncols=2)

  os.chdir(nb.workingdir)
  for nb.exp_path, stop in zip(findfolders_abspath(), exp_folders_to_skip):
    if stop:
      continue
    nb.exp_name = os.path.basename(nb.exp_path)
    """ important: setting working folder """
    os.chdir(nb.exp_path)
    """ defining the folder where plots will be saved """
    nb.suboutpath = os.path.join(nb.workingdir, nb.outpath, os.path.basename(nb.exp_path))
    if not os.path.exists(nb.suboutpath):
      os.makedirs(nb.suboutpath)

    """ reading or generate csv """
    retval = read_predata(nb)
    if retval == -1:
      print("read_predata went wrong. skipping this.")
      return -1

    """ parsing seq file to extract information and calculating the
    app. cryst. size to plot along with phase weight fractions """
    for nb.ref_path in findfolders_abspath():
      retval = read_refined_data(nb)
      # retval = plot_refined_data(nb)
    """ finding all folders (with a sequentially refined model)
    in current experimental folder """

    """ We try to combine all the piece-wise refinements """
    retval = prepare_piecewise_refinements(nb)
    retval = plot_piecewise_refinements(nb)

  """ checking if no plots have been generated. If not, output folder is deleted. """
  retval = clean_after_plotting(nb)
  return 0



def plot_piecewise_refinements(nb):

  """ plotting the concatenated _size and _PWF .csv found in refinement folders """
  # plot_seq_tools.plotphaseweightfractions(nb, title = nb.exp_name)
  # plot_seq_tools.size_plot(nb, title = nb.exp_name)

  """ (x,2) subplots for each xth experiments. """
  print(nb.TwoByTwo.flataxes)
  trace()
  mylength =len(nb.TwoByTwo.flataxes)
  nb.TwoByTwo.ax = nb.TwoByTwo.flataxes.pop(0)
  nb.TwoByTwo.title = "a test"+str(mylength);
  nb.TwoByTwo.hidelegend = False;  nb.TwoByTwo.xpositionLegend = None
  nb.TwoByTwo.plotphaseweightfractions()
  nb.TwoByTwo.ax = nb.TwoByTwo.flataxes.pop(0)
  nb.TwoByTwo.savefig = mylength <= 2
  print("nb.TwoByTwo.savefig =", nb.TwoByTwo.savefig)
  nb.TwoByTwo.title = "_nonaggr";
  nb.TwoByTwo.xpositionLegend = None;  nb.TwoByTwo.hidelegend = False
  nb.TwoByTwo.size_plot()

  """ all data plotted in (2,1) subplots """
  nb.TwoByOne.remove_xlabel_and_xticks = True;  nb.TwoByOne.ax = nb.TwoByOne.flataxes[0];
  nb.TwoByOne.title = "a test";  nb.TwoByOne.hidelegend = False;
  nb.TwoByOne.xpositionLegend = 0.5
  nb.TwoByOne.plotphaseweightfractions()

  nb.TwoByOne.savefig = nb.TwoByTwo.savefig

  nb.TwoByOne.remove_xlabel_and_xticks = False;  nb.TwoByOne.ax = nb.TwoByOne.flataxes[1]
  nb.TwoByOne.hidelegend = False;  nb.TwoByOne.xpositionLegend = 0.5;  nb.TwoByOne.title = "_aggr"
  nb.TwoByOne.size_plot()

  """ (2,2) subplot with reduced view """
  myxlim = (0,16)
  nb.TwoByTwoTruncated.remove_xlabel_and_xticks = True;  nb.TwoByTwoTruncated.ax = nb.TwoByTwoTruncated.flataxes[0]
  nb.TwoByTwoTruncated.title = "ThisTitleWillNotBeShown";  nb.TwoByTwoTruncated.hidelegend = True;  nb.TwoByTwoTruncated.xlim = None
  nb.TwoByTwoTruncated.plotphaseweightfractions()
  # nb.TwoByTwoTruncated._setting_up_legend( nb.TwoByTwoTruncated.flataxes[0], xposition = 1.1 )

  nb.TwoByTwoTruncated.ax = nb.TwoByTwoTruncated.flataxes[1];  nb.TwoByTwoTruncated.title = "ThisTitleWillNotBeShown"
  nb.TwoByTwoTruncated.hidelegend = False;  nb.TwoByTwoTruncated.xpositionLegend = -0.1;  nb.TwoByTwoTruncated.xlim = myxlim
  nb.TwoByTwoTruncated.plotphaseweightfractions()

  nb.TwoByTwoTruncated.remove_xlabel_and_xticks = False;  nb.TwoByTwoTruncated.ax = nb.TwoByTwoTruncated.flataxes[2]
  nb.TwoByTwoTruncated.title = "_aggr_trunc";  nb.TwoByTwoTruncated.hidelegend = True;  nb.TwoByTwoTruncated.xlim = None
  nb.TwoByTwoTruncated.size_plot()
  # nb.TwoByTwoTruncated.flataxes[1].legend()._visible = False

  nb.TwoByTwoTruncated.savefig = nb.TwoByTwo.savefig

  nb.TwoByTwoTruncated.ax = nb.TwoByTwoTruncated.flataxes[3];  nb.TwoByTwoTruncated.title = "_aggr_trunc"
  nb.TwoByTwoTruncated.hidelegend = False;  nb.TwoByTwoTruncated.xpositionLegend = -0.1;  nb.TwoByTwoTruncated.xlim = myxlim
  nb.TwoByTwoTruncated.size_plot()
  # nb.TwoByTwoTruncated.flataxes[3].legend()._visible = False


  return 0







def prepare_piecewise_refinements(nb):
  """ search for all _size and _PWF .csv in refinement folders
  to read in as pd.DataFrame and merge of possible.

  these df are stored in the notebook-object. """
  size_paths = []
  seq_paths = []
  phases = []
  for f in findfolders_abspath(nb.exp_path):
    if os.path.exists(os.path.join(f,"_size.csv")):
      size_paths += [ os.path.join(f,"_size.csv") ]
      seq_paths += [ os.path.join(f,"_seq.csv") ]
      pcr = sorted(findfiles_abspath(".pcr", path = f), key = lambda x : len(x) )[0]
      phases.append(extract_pcr.getphases(pcr))
  mysize = []
  myseq = []
  for csv in seq_paths:
    myseq.append( pd.read_csv(csv, index_col = "NEW_REFINEMENT") )
  for csv in size_paths:
    mysize.append( pd.read_csv(csv, index_col = "NEW_REFINEMENT") )

  # myseq = myseq
  # mysize = mysize
  unique_phase_nr2 = 999
  import itertools
  flat = list(itertools.chain.from_iterable(phases))
  flatdict = dict()
  for phase in set(flat):
    if phase in flatdict.keys():
      continue
    for num,phs in enumerate(phases):
      try:
        j = phs.index(phase) + 1
        # print("j =",j)
        # print("num =",num)
        # print("phases =",phs)
        # print("phase =",phase)
        # print("flatdict =",flatdict)
        # print("unique_phase_nr2 =",unique_phase_nr2)
      except ValueError:
        continue
      # print("mysize[num].columns =",mysize[num].columns)
      myseq[num].columns = [ re.sub("_ph"+str(j), "_ph"+str(unique_phase_nr2),x) for x in myseq[num].columns ]
      mysize[num].columns = [ re.sub("_ph"+str(j), "_ph"+str(unique_phase_nr2),x) for x in mysize[num].columns ]
      # print("mysize[num].columns =",mysize[num].columns)
      flatdict[phase] = unique_phase_nr2
    unique_phase_nr2 -= 1
  myseq = sorted(myseq, key = lambda x : min(x.index))
  mysize = sorted(mysize, key = lambda x : min(x.index))
  # for p in myseq: print(p.index)
  # for p in mysize: print(p.index)
  nb.seqdf = pd.concat(myseq)
  nb.sizedf = pd.concat(mysize)
  # print(flatdict)
  # for p in mysize: print(p.columns)
  keys = [k for k in flatdict.keys()]
  inv_map = {val: key for key, val in flatdict.items()}
  nb.phases = []
  for num,mykey in enumerate(sorted(inv_map.keys()), 1):
    phase = inv_map[mykey]
    nb.phases.append(phase)
    nb.seqdf.columns = (lambda mykey=mykey, nb = nb, num = num: [re.sub("_ph"+str(mykey), "_ph"+str(num),x) for x in nb.seqdf.columns])()
    nb.sizedf.columns = (lambda mykey=mykey, nb = nb, num = num: [re.sub("_ph"+str(mykey), "_ph"+str(num),x) for x in nb.sizedf.columns])()

  nb.lorsizelist = [col for col in nb.seqdf.columns if col.startswith("Y-cos") and not col.endswith("_err")]

  """ we use the extracted time stamps from the nb.datdf.columns
  for the concatenated dataframes. We blindly assume that the
  first dataframe is added manually to give the zero point of
  the experiment.
  This code is only useful for neutron data!

  extra code has to be written for synchrotron data. """
  # print("nb.seqdf.columns =",nb.seqdf.columns)
  # print("nb.sizedf.columns =",nb.sizedf.columns)

  if nb.datdftype == "dmc":
    """ we skip the first frame for neutron data.
    this will be changed in the future. the first frame will be indexed 1 and not 0
    in this way it can be included in the sequential refinement. """
    try:
      nb.seqdf.index = nb.datdf.columns[1:]
    except:
      print("len(nb.seqdf.index) =", len(nb.seqdf.index))
      print("len(nb.datdf.columns[1:]) =", len(nb.datdf.columns[1:]) )
      print("seq_paths[0] =\n", seq_paths[0])
      trace()
    nb.sizedf.index = nb.seqdf.index
  elif nb.datdftype == "dioptas":
    FirstFrame = int(nb.datdf.columns[0].strip("frame"))
    nb.seqdf.index = [(int(x.strip("frame"))-FirstFrame) * 5 / 60 for x in nb.datdf.columns]
    nb.sizedf.index = nb.seqdf.index

  nb.seqdf.index.name = "time [minutes]"
  nb.sizedf.index.name = nb.seqdf.index.name
  # nb.seqdf = sorted(myseq, key = lambda x : min(x.index))
  # nb.sizedf = sorted(mysize, key = lambda x : min(x.index))
  return 0




def clean_output_folder(nb):
  """ removing everything in the destination folder for graphs.
  This is maybe not necessary, but it does ensure that old/obsolete
  refinements are removed. """
  try:
    sys.stdout.write("... \"_mytest\" folder")
    shutil.rmtree(nb.outpath)
    sys.stdout.write(" - SUCCESFULLY REMOVED!\n")
    return 0
  except:
    sys.stdout.write(" - FAILURE!\n")
    return -1

def clean_after_plotting(nb):
  os.chdir(nb.outpath)
  for f in findfolders_abspath():
    if os.listdir(f) == []:
      os.rmdir(f)
      print("    removed " + str(os.path.basename(f)) + " from " + str(os.path.dirname(nb.outpath)) )
    # return # premature return to save time when debugging.
  return 0


def read_predata(nb):
  """ reading in or generating data frames """
  datcsv = [x for x in os.listdir() if x.endswith("_dats.csv")]
  if len(datcsv) > 1:
    print("found more _dats.csv, we skip workingdir =",nb.workingdir)
    return -1
  if len(datcsv) == 1:
    nb.datdf = pd.read_csv(datcsv[0])
    if "angle" in nb.datdf.columns:
      nb.datdf = nb.datdf.set_index("angle")
    try:
      """ when reading in dataframe, the columns are strings,
      we need them to be float64 """
      """ this is only needed for DMC data... can we homogenize the dataframes?
      EDIT: we want this inhomogenity now to decide what type datdf is :) """
      nb.datdf.columns = [float(x) for x in nb.datdf.columns]
      nb.datdftype = "dmc"
    except:
      nb.datdftype = "dioptas"
  else:
    datafolders = [x[0] for x in os.walk(nb.exp_path) if x[0].split("\\")[-1] == "tset"]
    if len(datafolders) == 0:
      return -1
    if len(datafolders) != 1:
      print("There are more than one subfolder named 'tset'. skipping ", nb.exp_path)
      return -1
    try:
      nb.datdf = datfiles2datcsv.generate_CSV_from_dmc_dat(workingdir = datafolders[0])
      nb.datdftype = "dmc"
    except Exception as dmc_error:
      try:
        nb.datdf = datfiles2datcsv.genereate_CSV_from_dioptas_dat(workingdir = datafolders[0])
        nb.datdftype = "dioptas"
      except Exception as dioptas_error:
        print("dmc_error =",dmc_error)
        print("dioptas_error =",dioptas_error)
        import pdb; pdb.set_trace();
        print("get creative.")
        return -1
  return 0

def plot_predata_handler(nb):
  """ missing introduction """
  print("--inside experiment folder:", os.path.basename(nb.exp_path))

  """ plotting data frames """
  """ these try except statements should be changed to if os.path.exsists'ish """
  if nb.datdftype == "dioptas":
    d1 = plot_predata.Dioptas(nb, minute_range = (0,24), printsize = "four_in_docx")
    retval1 = d1.run()
    print("found and plotted dioptas .dat/_dats_csv file(s)")
  elif nb.datdftype == "dmc":
    d2 = plot_predata.Dmc(nb, minute_range = (0,230), printsize = "four_in_docx")
    retval2 = d2.run()
    print("found and plotted dmc .dat/_dats_csv file(s)")
  else:
    print("found neither dioptas or DMC data to make heat map plots.")
    return -1
  """ no mistakes """
  return 0




def read_refined_data(nb):
  """  """
  print("    working on:", nb.ref_path)
  """ important: setting new working directory """
  os.chdir(nb.ref_path)
  """ defining object to store information passed between functions. """
  seq = findfile_abspath(".seq")
  """ find the pcr with shortest filename """
  pcr = findfiles(".pcr")
  if len(pcr) == 0: print("!-!-!-! skipping \"" + nb.exp_path + "\" + --> no .pcr was found"); return -1
  pcr.sort(key=lambda x: len(x))
  pcr = pcr[0]

  """ extracting list with phasenames """
  nb.phases = extract_pcr.getphases(pcr)

  """ extracting wavelength """
  nb.wavelength = extract_pcr.getwavelength(pcr)

  """ storing the folder name of the sequentially refined model """
  nb.ref_name = os.path.basename(os.getcwd())

  """ check if _seq.csv exists and if it has been changed later than its
  parent .pcr file. This saves some computing power and man-hour.
  """
  seqcsvpath = os.path.join(nb.ref_path, "_seq.csv")
  if os.path.exists(seqcsvpath) and os.stat(seqcsvpath).st_mtime > os.stat(pcr).st_mtime:
    """ st_mtime is the time in seconds since 1970ish telling
    when the content of file was last changed.
    with st_mtime we ask if the pcr was modified before the seqcsv file

    26. marts 2018: I, Pelle, was checking if this was actually executing as expecting. Both _size and _seq csv files
    were updated as intended! """
    nb.seqdf = pd.read_csv(seqcsvpath, index_col = "NEW_REFINEMENT")
  else:
    seqcsvpath = seqfile2seqcsv.seq2csv(seq)
  nb.seqdf = pd.read_csv(seqcsvpath, index_col = "NEW_REFINEMENT")

  """ finding the lorentizian size broadening parameters used for
  calculating the app. cryst. sizes """
  nb.lorsizelist = [col for col in nb.seqdf.columns if col.startswith("Y-cos") and not col.endswith("_err")]

  """ check if _size.csv exists and if it has been changed later than its
  parent .pcr file. This saves some computing power and man-hour.
  """
  sizecsvpath = os.path.join(nb.ref_path, "_size.csv")
  if os.path.exists(sizecsvpath) and os.stat(sizecsvpath).st_mtime > os.stat(pcr).st_mtime:
    """ st_mtime yields the number of seconds since 1970'ish """
    nb.sizedf = pd.read_csv(sizecsvpath, index_col = "NEW_REFINEMENT")
  else:
    nb.sizedf = seqcsv2sizecsv.calculate_size(nb)




def plot_refined_data(nb):
  """ plotting refined data """

  os.chdir(nb.ref_path)

  """ plotting the phase weight fractions """
  plot_seq_tools.plotphaseweightfractions(nb)

  """ plotting and saving the app. cryst. size informations. """
  plot_seq_tools.size_plot(nb)

  return 0






# END



# find_tools.py

import os
from wrappers import RestorePath # this does not work as intended!

# @RestorePath
def findfile(extension, path = ""):
    """
    Returns name of file with extension in current working directory

    Args:
        extension: Examples include ".txt", ".pcr", ".py" etc.

    Raises:
        No exception is raised for multiple files with file ending "extension".
        but comments are printed to the console about the nature of extension
        and how many additional files were found.

    """
    path = os.getcwd() if path == "" else path
    files = [f for f in os.listdir( path ) if f.endswith( extension )]
    searched_file = files[0] if len(files) == 1 else "error!"
    # if searched_file == "error!":
    #     print("error in folder:", path)
    #     print("number of files with extension", "\"" + extension + "\"", "is", len( files ))
    return searched_file

# @RestorePath
def findfile_abspath(extension, path = ""):
  """ returns only one file with extension """
  path = os.getcwd() if path == "" else path
  files = [os.path.join( path, f ) for f in os.listdir( path ) if f.endswith( extension )]
  searched_file = files[0] if len(files)==1 else "error!"
  # if searched_file == "error!":
  #   print("error finding file with extension:", extension, "in folder", path )
  #   print("number of files with extension:", len( files ))
  return searched_file

# @RestorePath
def findfiles(extension, path = ""):
    """
    Returns all files with specified extension in the current working directory

    Args:
        extension: Examples include ".txt", ".pcr", ".py" etc.

    Returns:
        list of files with specified extension.

    """
    path = os.getcwd() if path == "" else path
    return [f for f in os.listdir( path ) if f.endswith( extension )]

# @RestorePath
def findfiles_abspath(extension, path = ""):
  """ returns all files with specified extension """
  path = os.getcwd() if path == "" else path
  files = [os.path.join( path, f ) for f in os.listdir( path ) if f.endswith( extension )]
  return files

# @RestorePath
def findfolders(path = "" ):
    """
    Returns all folders in current working directory

    Returns:
        list of directories in current working directory

    """
    path = os.getcwd() if path == "" else path
    folders = [f for f in os.listdir( path ) if os.path.isdir( os.path.join(path, f) )]
    return folders

# @RestorePath
def findfolders_abspath( path = "" ):
  """ returns all folders in cwd """
  path = os.getcwd() if path == "" else path
  """ if current work dir is not path, then previous code fails. Needs to change dir... """
  folders = [os.path.join( path, f ) for f in os.listdir( path ) if os.path.isdir( os.path.join(path, f) )]
  return folders



# plot_seq_tools.py

import os, pandas as pd, matplotlib.pyplot as plt, numpy as np, re
from auxiliary import cm2inch
from itertools import chain
import pdb; trace = pdb.set_trace

""" choosing plotting style """
# plt.style.use("seaborn-darkgrid") # plt.style.available
plt.style.use("classic") # plt.style.available


class SubplotManager():
  def __init__(self, notebook, nrows = 1, ncols = 1,sharex=False, sharey=False, squeeze=True, subplot_kw=None, gridspec_kw=None, figsize = (8,8), **fig_kw):
    self.fig, self.ManyAx = plt.subplots(
    nrows=nrows,
    ncols=ncols,
    sharex=sharex,
    sharey=sharey,
    squeeze=squeeze,
    subplot_kw=subplot_kw,
    gridspec_kw=gridspec_kw,
    figsize = figsize,
    **fig_kw,
    )
    # plt.subplots_adjust(left=1, bottom=1, right=1, top=1, wspace=1, hspace=1)
    self.shape = (nrows, ncols)
    self.size = nrows*ncols
    self.savefig = False # save plotting preliminiary subplots until we are through all subfolders.

    self.flataxes = list(chain(*self.ManyAx)) if self.size > 2 else list(chain(self.ManyAx))
    self.ax = ""
    self.title = ""
    self.xpositionLegend = 0.5
    self.xlim = None
    self.hidelegend = False

    self.notebook = notebook
    self.notebook._textsize = 10
    self.notebook._ticksize = 7
    self.remove_xlabel_and_xticks = False
    self.dpi = 250

    """ this is to modify the legend size... """
    import matplotlib
    matplotlib.rcParams.update({'font.size': self.notebook._textsize})


  # def show(self):
  #   plt.show()


  def _reduce_displayed_points(self, df):
    """ reducing the number of displayed datapoints to X
    for neutron data this removes all kinks and is not preferable. """
    if self.notebook.datdftype == "dioptas":
      number_of_points_displayed = 100#(notebook.seqdf.shape[0]-1) // 10
    elif self.notebook.datdftype == "dmc":
      number_of_points_displayed = df.shape[0]-1
    view = [int(i) for i in np.linspace(0, df.shape[0]-1, number_of_points_displayed)]
    view = set(view)
    view = sorted(list(view))
    view = df.iloc[view]
    return view

  def _get_nonmagnetic_phases(self):
    return [phase for phase in self.notebook.phases if not " magnetic" in phase]


  def _label_text(self, mystring, label):
    """ returns a latexified string to be used in plt.
    e.g. 500oC --> 500$^o$C """
    trans1 = re.sub("(\d)oC","\g<1>°C",mystring)
    trans2 = re.sub(" nuclear","",label)
    return trans1 + ": " + trans2



  def _add_subplot(self, df, y = "y", yerr = "yerr", label = "label"):
    if self.ax == "":
      self.ax = df.plot(y=y,yerr=yerr, label = label, figsize = cm2inch(8,6))
    else:
      df.plot(y=y,yerr=yerr, ax = self.ax, label = self._label_text(self.notebook.exp_name, label) )


  def _setting_up_plot(self, ylabel_text="ylabel_text"):
    """ handling graph text """
    # ax.autoscale_view('tight')


    """ ylabel text and size """
    self.ax.set_ylabel(ylabel_text, size = self.notebook._textsize)
    # title = self.notebook.ref_name if title == "" else title
    # title = ""
    # ax.set_title(title, size = self.notebook._textsize)
    for tick in self.ax.xaxis.get_major_ticks():
      tick.label.set_fontsize(self.notebook._ticksize)
    for tick in self.ax.yaxis.get_major_ticks():
      tick.label.set_fontsize(self.notebook._ticksize)

    if self.remove_xlabel_and_xticks:
      ticklabels = [label for label in self.ax.xaxis.get_ticklabels()]
      self.ax.xaxis.set_ticklabels(["" for x in ticklabels])
      # self.ax.get_xaxis().get_major_ticks._visible = False
      self.ax.set_xlabel("", size = self.notebook._textsize)
    else:
      self.ax.set_xlabel(self.notebook.seqdf.index.name, size = self.notebook._textsize)

    """ ensure output directory exists """
    if not os.path.exists(self.notebook.suboutpath):
      os.makedirs(self.notebook.suboutpath)

    """ adjusting ylim """
    self.ax.set_ylim([-5,125])


    self.ax.grid()

    self._setting_up_legend()

  def _setting_up_legend(self):
    """ adding background color to legends """
    mytext = self.ax.legend().get_texts()
    legend = self.ax.legend(frameon = 1, loc = "best", prop = {"size": self.notebook._ticksize} )
    frame = legend.get_frame()
    frame.set_facecolor('white')
    frame.set_edgecolor('black')
    if self.xpositionLegend == None:
      # ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15),
      #       ncol=2, fancybox=True, shadow=True)
      pass
    elif len(mytext) > 4:
      self.ax.legend(loc='upper center', bbox_to_anchor=(self.xpositionLegend, 1.15),
      ncol=3, fancybox=True, shadow=True)
    elif len(mytext) > 2:
      self.ax.legend(loc='upper center', bbox_to_anchor=(self.xpositionLegend, 1.15),
      ncol=2, fancybox=True, shadow=True)
      """ this is to modify the legend size... """
    import matplotlib
    matplotlib.rcParams.update({'font.size': self.notebook._ticksize})
    if self.hidelegend:
      self.ax.legend()._visible = False




  def plotphaseweightfractions(self):
    """ plotting the phase weight fractions from seqdf for all phases in refined model """

    """ defining path to save plot """
    outpath = self.notebook.suboutpath if self.ax == "" else self.notebook.outpath

    def get_cols_to_plot(df):
      """ making a list of weight_fractions for all phases.

      this became a subfunctions in order to skip it when debugging..."""
      cols_to_plot = []
      """ a bit tricky construct. we are working with both magnetic and nuclear
      phases. I can only assume that the FP user has bundled all open parameters
      to one phase (either nuclear or magnetic). We already have a list of
      lorentsian size distributions. we assume these belong to same phase as the
      refined intensity / weight_fraction """
      i = 0
      for col in df.columns:
        if i < len(self.notebook.lorsizelist) and "Weight_Fraction" in col and not col.endswith("_err"):
          # print(col.lstrip("Weight_Fraction"))
          # print(self.notebook.lorsizelist[i])
          if col.lstrip("Weight_Fraction") in self.notebook.lorsizelist[i]:
            cols_to_plot.append(col)
            i+=1
      return cols_to_plot


    """ organising reduced dataframe (view), the phase names and the columns to plot (ignoring magnetic phases) """
    view = self._reduce_displayed_points(self.notebook.seqdf)
    phases = self._get_nonmagnetic_phases()
    cols_to_plot = get_cols_to_plot(self.notebook.seqdf)

    """ plotting phase weight fraction for all nonmagnetic phases in model """
    for phase, col in zip(phases, cols_to_plot):
      # self._add_subplot(view, ax, y = col, yerr = col+"_err", label = phase)
      view.plot(y=col,yerr=col+"_err", ax = self.ax, label = self._label_text(self.notebook.exp_name, phase) )


    """ handling x, y, title labels and adjusting graphs etc.. """
    self._setting_up_plot(ylabel_text = "phase weight fraction [%]")

    if self.xlim != None:
      self.ax.set_xlim(self.xlim)

    # plt.tight_layout()
    """ saving graph """
    if self.savefig:
      trace()
      plt.tight_layout()
      self.fig.savefig(os.path.join(outpath, "PVF_"+self.title+".png"), dpi = self.dpi)
    #
    # """ avoid memory leak """
    # plt.close()
    return



  def size_plot(self):
    """ plotting the app. cryst. sizes found in sizedf """

    """ defining path to save plot """
    outpath = self.notebook.suboutpath if self.ax == "" else self.notebook.outpath

    """ check if sizedf is empty """
    if self.notebook.sizedf.size == 0:
      message = "self.notebook.sizedf.size = " + str(self.notebook.sizedf.size)
      with open(os.path.join(self.notebook.suboutpath,self.notebook.ref_name+".txt"), "w") as writing:
        writing.write(message)
      print("        " + message)
      return

    """ reducing the number of points displayed... """
    view = self._reduce_displayed_points(self.notebook.sizedf)

    """ plotting app. cryst. sizes for all phases in refined model """
    for phase, col in zip(self._get_nonmagnetic_phases(), self.notebook.lorsizelist):
      target = "ab-size" + col.strip("Y-cos") # Y-cos is lorentzian size broadening
      # print("target =",target)
      self._add_subplot(view, y = target, yerr = target+"_err", label = phase)

    self._setting_up_plot(ylabel_text = "app. cryst. size [nm]")


    if self.xlim != None:
      self.ax.set_xlim(self.xlim)

    # """ ensuring every text is within plot """
    """ saving plot in ./root/_mytest """
    if self.savefig:
      plt.tight_layout()
      self.fig.savefig(os.path.join(outpath,"subpl"+self.title+".png"), dpi = self.dpi)

    # """ avoiding memory leak """
    # plt.close()
    return









# end



# plot_seq_tools.py

import pandas as pd, numpy as np, math
# from plot_seq_tools import Notebook # amazingly we don't need to import Notebook!

""" global variable... """
upper_limit_in_nm = 120

def calculate_size(notebook):
  """
  df:
    pandas.dataframe with columns of refined parameters extracted from FullProf's .seq file
  wavelength:
    floating point number in Ångstrøm.

  What is calculated:
    Extracted from Anna's Excel Sheet. Do not change!

    CITATIONS MISSING!

    Y[°]	from .out as "Y-cos_ph1_pat1"

    σ_Y[°]	from .out error for above

    S_z[Å^-1]	from .out as "L-Size_ph1_pat1"

    σ_sz[Å^-1]	from .out error for above.

    Y[Å^-1] = Y[°] * PI^2 / 180 / 2 / wavelength[Å] * 1000

    σ_Y[Å^-1] = σ_Y[°] * PI^2 / 180 / 2 / wavelength * 1000

    # ab size

    a,b-size[nm] = 100 / Y[Å^-1]

    σ_a,b-size[nm] = 100 * (σ_Y[Å^-1]^2 / Y[Å^-1]^4)^0.5

    # c size

    c-size[nm] = 100 / (Y[Å^-1] + S_Z[Å^-1])

    σ_c-size[nm] = =100 * ((σ_Y[Å^-1]^2 + σ_sz[Å^-1]^2) / (Y[Å^-1] + S_z[Å^-1])^4)^0.5

  """
  def calc_inverse_lorentzian(ydegree):
    """ helper function (can also be used for ydegree_error!)

    Y[Å^-1] = Y[°] * PI^2 / 180 / 2 / wavelength[Å] * 1000

    this function only works because <wavelength> is already defined in the parent function

    """
    yinverseangstroem = ydegree * math.pow(math.pi, 2) / 36.0 / notebook.wavelength * 100
    return yinverseangstroem

  def multiply_invers_100(somevalue):
    """ helper function

    a,b-size[nm] = 100 / Y[Å^-1]

    """
    try:
      absize = 100 / somevalue
    except ZeroDivisionError:
      """ we return 100 nm. This is well beyound the reoslution limitation
      of DMC (reliable below 40 nm) and also for synchrotrons such as P02.1
      beamline at Petra III (resolution limit below ~ 100 nm) """
      global upper_limit_in_nm
      absize = upper_limit_in_nm
    return absize


  columnnames = []

  for col in notebook.lorsizelist:
    """ working with Y-cos_ph1_pat1
    to create L-Size_ph1_pat1, ab-size and c-size
    plus all errors """
    lorsize = "Y-cos"
    columnnames.append(col)
    columnnames.append(col + "_err")
    stripped = col.strip(lorsize)
    lorsizedistribution = "L-Size"
    columnnames.append(lorsizedistribution+stripped)
    columnnames.append(lorsizedistribution+stripped+"_err")
    absize = "ab-size"
    columnnames.append(absize+stripped)
    columnnames.append(absize+stripped+"_err")
    csize = "c-size"
    columnnames.append(csize+stripped)
    columnnames.append(csize+stripped+"_err")

  """ creating the outlay for the sizedf filled with zeros """
  myzeros = np.zeros( shape=(notebook.seqdf[notebook.lorsizelist[0]].size, 8*len(notebook.lorsizelist)) )

  """ initializing sizedf with columns from columnnames"""
  newdf = pd.DataFrame(myzeros, index = notebook.seqdf.index, columns = columnnames)
  for col in newdf.columns:
    try:
      """ importing columns from seqdf to sizedf """
      newdf[col] = notebook.seqdf[col]
    except KeyError:
      pass
  for col in columnnames:
    """ filling out the columns corresponding to ab- and c-size """
    if lorsize in col and not col.endswith("_err"):
      newdf[col] = newdf[col].apply(calc_inverse_lorentzian)
      newdf[col+"_err"] = newdf[col+"_err"].apply(calc_inverse_lorentzian)
      stripped = col.strip(lorsize)
      newdf[absize+stripped] = newdf[col].apply(multiply_invers_100)
      """ σ_a,b-size[nm] = 100 * (σ_Y[Å^-1]^2 / Y[Å^-1]^4)^0.5 """
      target = absize+stripped+"_err"
      newdf[target] = newdf[col+"_err"].apply(lambda x: math.pow(x,2))
      newdf[target] = newdf[target].div(newdf[col].apply(lambda x: math.pow(x,4)))
      newdf[target] = newdf[target].apply(lambda x: math.sqrt(x))
      newdf[target] = newdf[target].apply(lambda x: x*100)
      """ c-size[nm] = 100 / (Y[Å^-1] + S_Z[Å^-1]) """
      target = csize+stripped
      newdf[target] = newdf[col] + newdf[lorsizedistribution+stripped]
      newdf[target] = newdf[target].apply(multiply_invers_100)
      """ σ_c-size[nm] = =100 * ((σ_Y[Å^-1]^2 + σ_sz[Å^-1]^2) / (Y[Å^-1] + S_z[Å^-1])^4)^0.5 """
      target = csize+stripped+"_err"
      # if "_ph3" in col:
      #   print("None =",None)
      #   import pdb; pdb.set_trace();
      newdf[target] = newdf[col+"_err"].apply(lambda x: math.pow(x,2))
      newdf[target] = newdf[target] \
        + newdf[lorsizedistribution+stripped+"_err"].apply(lambda x: math.pow(x,2))
      newdf[target] = newdf[target] \
        .div((newdf[col] + newdf[lorsizedistribution+stripped]).apply(lambda x: math.pow(x,4)))
      newdf[target] = newdf[target].apply(lambda x: math.sqrt(x))
      newdf[target] = newdf[target].apply(lambda x: x*100)
      """ some NaNs may have been introduced, we replace by 100 (nm) that marks
      the upper insturmental resolution for both neutron and synchrotron diffraction

      EDIT: infinity was introduced in calculating error. These infinity are
      replaced with 100 as well."""
      # celebrate!
  global upper_limit_in_nm
  newdf.fillna(upper_limit_in_nm,inplace=True)
  newdf = newdf.replace(np.inf, upper_limit_in_nm)
  for col in columnnames:
    if col.endswith("_err"):
      continue
    indices_to_change = np.where(newdf[col] >= upper_limit_in_nm)[0]
    if indices_to_change.size != 0:
      newdf[col][newdf.index[indices_to_change]] = upper_limit_in_nm
      newdf[col+"_err"][newdf.index[indices_to_change]] = 0


  # print("notebook.exp_name =",notebook.exp_name)
  # import pdb; pdb.set_trace();

  """ cleaning up if NaNs have been introduced by the multiply_invers_100 function

  EDIT: cleaning up (that is removing all entries with NaN) caused the sizedf
  to be empty, even though one phase had a defined calculated size. But a second
  phase without a broadening could hide this fact. Instead of NaN, a 200 nm is appended
  in the multiply_invers_100 function.

  The code below has not been changed... should still work. """
  indices_with_nan = set(np.where(np.isnan(newdf))[0])
  indices_to_keep  = [index for index in range(newdf.index.size) if index not in indices_with_nan]
  newdf = newdf.iloc[indices_to_keep]
  """ saving to .csv """
  newdf.to_csv("_size.csv")
  """ return the sizedf dataframe """
  return newdf







# end



from prf_plot_tools import excluded_regions
import pandas as pd, numpy as np
import re

def read_prf(prfpath = ""):
  """ determines the type of the requested prf, reads it in accordingly and
  returns a tuple of prf_type and pd.dataframe

  THIS IS THE MAIN FUNCTION OF THIS COLLECTION
  """
  if prfpath == "": return ("NO PRF PATH GIVEN", -1)
  prf_type = get_prf_type(prfpath)
  if prf_type == "dmc":
    df = read_dmc(prfpath)
  elif prf_type == "xray":
    df = read_xray(prfpath)
  else:
    return ("NO PRF READ!", -1)
  return (prf_type, df)

def get_prf_type(prfpath):
  """ searching prf file until indicators for a specific diffraction data type is found.

  prf supported:
    dmc (and also HRPT)
    xray (and also dioptas generated 1-line patterns from synchrotrons)
  """
  with open(prfpath, "r") as reading:
    for num,line in enumerate(reading):
      is_dmc = re.search(".*92.*12.800.*0.10000", line)
      is_xray = re.search(".*2Theta.*Yobs.*Ycal.*Yobs-Ycal.*Backg.*Posr.*(hkl).*K",line)
      if is_dmc:
        return "dmc"
      elif is_xray:
        return "xray"
  return "no known prf-type was found..."


def read_dmc(prfpath):
  """ reading dmc/hrpt prf file and returns a pd.dataframe """
  with open(prfpath, "r") as reading:
    for num,line in enumerate(reading):
      if num >= 5:
        numbers = [float(number) for number in line.strip().split() ]
        if len(yobs) != entries:
          yobs += numbers
        elif len(ycalc) != entries:
          ycalc += numbers
        else:
          bragg_reflections += numbers
      elif num == 2:
        end,start,step = [float(number) for number in line.strip().split()[0:3]]
        angles = np.arange(start,end+step,step)
        entries = len(angles)
        yobs = list()
        ycalc = list()
        bragg_reflections = list()

  """ preparing dataframe """
  df = pd.DataFrame()
  for col_name, col in zip(["2Theta", "Yobs", "Ycal"], [angles, yobs, ycalc]):
    df[col_name] = col
  return df

def read_xray(prfpath):
  """ reading convential xray/synchroton prf file and returns a pd.dataframe """
  if prfpath == "": return "no valid prf was given"
  nb_excluded_regions = excluded_regions()
  df = pd.read_csv(prfpath, skiprows = 3+nb_excluded_regions, delimiter="[\t]+", engine="python")
  df = df[["2Theta","Yobs","Ycal"]]
  for col in df.columns.values:
    df[col] = pd.to_numeric(df[col], errors="coerce")
  df = df.dropna(axis=0)
  return df



# wrappers.py

import sys,os
from datetime import datetime
from functools import wraps

def RestorePath(func):
  """ takes cwd, executes the function, restores to the original cwd
  useful for subfunctions that might change the current working directory. """
  @wraps
  def wrapper(*args, **kwargs):
    oldpath = os.getcwd()
    res = func(*args,**kwargs)
    os.chdir(oldpath)
    return res
  return wrapper

def time_response(func, depth = ""):
    """
    use as a decoration to measure the execution time of the function call.
    """
    @wraps
    def function_wrapper(*args, **kwargs):
        t1 = datetime.now()
        print(depth + "#### " + func.__name__ + " ####")
        depth2 = depth + "    "
        res = func(*args, **kwargs)
        # print(res)
        t2 = datetime.now()
        print(depth + func.__name__ + " took " + str(t2-t1), " (hh:mm:ss.decimals)")
        return res
    return function_wrapper

class TraceCalls(object):
    """
    Use as a decorator on functions that should be traced. Several
    functions can be decorated - they will all be indented according
    to their call depth.
    """
    def __init__(self, stream=sys.stdout, indent_step=4, show_ret=False):
        self.stream = stream
        self.indent_step = indent_step
        self.show_ret = show_ret

        # This is a class attribute since we want to share the indentation
        # level between different traced functions, in case they call
        # each other.
        TraceCalls.cur_indent = 0

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            indent = '-' * TraceCalls.cur_indent
            argstr = ', '.join(
                #[repr(a) for a in args]) #+
                ["%s" % a for a, b in kwargs.items()])
            self.stream.write('%s> %s(%s)\n' % (indent, fn.__name__, argstr))

            TraceCalls.cur_indent += self.indent_step
            ret = fn(*args, **kwargs)
            TraceCalls.cur_indent -= self.indent_step

            if self.show_ret:
                self.stream.write('%s--> %s\n' % (indent, ret))
            return ret
        return wrapper



# setup.py

from distutils.core import setup
import os


setup(
    name='magnetmatter',
    version='0.2.5',
    packages=['magnetmatter',
      os.path.join('magnetmatter','modules'),
      os.path.join('magnetmatter','modules', '_plot_prf'),
      os.path.join('magnetmatter','modules','_plot_seq'),
      ],
    license='MIT',
    author='Pelle Garbus',
    author_email='garbus@inano.au.dk',
    description='Visualization of FullProf-refined neutron and X-ray powder diffraction data',
    long_description=open('README.txt').read(),
    install_requires=["numpy","pandas","matplotlib"],
    url = 'https://github.com/pgarbus/magnetmatter',
)



# prf_plot_tools.py

from decimal import Decimal
import re, os, math, numpy as np
from wrappers import TraceCalls
from find_tools import findfile, findfiles
import matplotlib.pyplot as plt
from auxiliary import cm2inch








@TraceCalls()
def get_spacegroups():
  """
  Creates a dictionary with spacegroups from FullProf .pcr file

  Returns:
    Dictionary of all spacegroups found bundled with a numeric key
    that correspond to the spacegroup's placement in the .pcr file.

  """
  pcr = findfile(".pcr")
  spacegroups = dict()
  with open(pcr,"r") as reading:
    i = 1
    for line in reading:
      if "<--Space group symbol" in line:
        spacegroup = line.split("<--Space group symbol")[0].strip()
        spacegroups[str(i)]=spacegroup
        i+=1
  return spacegroups

def excluded_regions():
  """
  Searches Fullprof .pcr file for code indicating exclusion of diffraction pattern.

  Returns:
    integer number of excluded regions.
    "0" = no excluded regions
    "1" = one excluded region
    etc.

  """
  pcr = findfile(".pcr")
  nb_excluded_regions = 0
  with open(pcr,"r") as reading:
    for line in reading:
      if "Excluded regions (LowT  HighT) for Pattern#" in line:
        while True:
          if reading.readline()[0] != "!": nb_excluded_regions += 1
          else: break
  return nb_excluded_regions



@TraceCalls()
def extract_out_file():
  """
  Extracts wavelengths, phase fractions, phases and refined parameters from FullProf .out file

  Returns:
    A tuple with the following 4 objects

    wavelength:
      floating point number of wavelength. The possible K alpha1
      and K alpha2 ratio is taken into account.

    frac_info: Dictionary of phase weight fractions.
      keys: phase name extracted in .out file.
      values: tuple of value and its error.

    phases: Dictionary of phases found in .out file bundled with numeric
            integer indication phase relationship.
      keys: phase number extracted from .out file.
      values: string of phase name as writting in .pcr file.

    refined_par: Dictionary of refined parameters.
      keys: FP notation of refined parameters.
      values: tupe with value and its error.

  """

  out = findfile(".out")
  with open(out,"r") as reading:
    """ extracting the relevant parameters """
    refined_par = dict()
    phases = dict()
    frac_info = dict()
    read_phase = False
    start = False; found = False
    for line in reading:
      if "Conventional Rietveld Rp,Rwp,Re and Chi2:" in line: #  => Conventional Rietveld Rp,Rwp,Re and Chi2:   13.1      16.9      6.06      7.765
        line = line.strip().split()
        rwp, chi2 = line[-3::2]
        refined_par["R wp / chi2"] = (float(rwp), float(chi2))
        continue
      if "Fract(%):" in line:
        line = line.strip().split("Fract(%):")[-1]
        #   => Bragg R-factor:   28.0       Vol:  591.768( 0.238)  Fract(%):   11.41( 2.25)
        frac, frac_error = [number.strip() for number in line.strip(")").split("(")]
        #  => Phase:  1     $\gamma$-Fe$_2$O$_3$ nuclear
        phasename = re.search("=> Phase:\s*\d+\s*([^\s].+)", prev_line.strip()).group(1)
        frac_error = "100.00" if frac_error == "*****" else frac_error
        frac_info[phasename] = (float(frac), float(frac_error))
      prev_line = line

      if "Wavelengths:" in line:
        w1, w2 = line.strip().split()[-2:]
      if "Alpha2/Alpha1 ratio:" in line:
        ratio = line.strip().split()[-1]
        wavelength = (float(w1) + float(w2)*float(ratio)) / (1+float(ratio))
      if read_phase:
        phases[phase_nr] = line.strip()
        read_phase = False
      if "=> Phase  No." in line:
        phase_nr = line.strip().split()[-1]
        read_phase = True
      if not start:
        start = re.search("SYMBOLIC NAMES AND FINAL VALUES AND SIGMA OF REFINED PARAMETERS:",line)
      if start:
        if re.search("Number of bytes for floating point variables",line):
          break
        """       ->  Parameter number    7 :      L-Size_ph1_pat1     1.4556953    ( +/-     1.4023348     ) """
        found = re.search(":\s*([^\s]+)\s*([^\s]+)\s*\( \+/-\s*([^\s]+)\s*\)",line)
      if found:
        key,value,error = found.group(1),found.group(2),found.group(3)
        refined_par[key] = (float(value), float(error))
  return wavelength, frac_info, phases, refined_par

@TraceCalls()
def calculate_size(wavelength = "", phases = "", refined_par = ""):
  """
  returns a dictionary containing ab- and c-size for each phase calculated from Y-cos and L-size.

  Args:
    phases: Dictionary
      keys: index
      value: string of phasename
    refined_par: Dictionary
      key: FP name of refined parameter
      value: tuple(value, error)
  Returns:
    dictionary:
      key: phase
      value:
        dict(key=ab- or c-size, value = tuple(value, error))

  What is calculated:
    Extracted from Anna's Excel Sheet. Do not change!

    CITATIONS MISSING!

    Y[°]	from .out as "Y-cos_ph1_pat1"

    σ_Y[°]	from .out error for above

    S_z[Å^-1]	from .out as "L-Size_ph1_pat1"

    σ_sz[Å^-1]	from .out error for above.

    Y[Å^-1] = Y[°] * PI^2 / 180 / 2 / wavelength[Å] * 1000

    σ_Y[Å^-1] = σ_Y[°] * PI^2 / 180 / 2 / wavelength * 1000

    # ab size

    a,b-size[nm] = 100 / Y[Å^-1]

    σ_a,b-size[nm] = 100 * (σ_Y[Å^-1]^2 / Y[Å^-1]^4)^0.5

    # c size

    c-size[nm] = 100 / (Y[Å^-1] + S_Z[Å^-1])

    σ_c-size[nm] = =100 * ((σ_Y[Å^-1]^2 + σ_sz[Å^-1]^2) / (Y[Å^-1] + S_z[Å^-1])^4)^0.5

  """
  res_dict = dict()
  ycos = list()
  gsize = list()
  for key in refined_par.keys():
    if "Y-cos" in key:
      ycos.append(key)
    elif "G-Size" in key:
      gsize.append(key)
  for key in ycos:
    phase_pattern = re.search("_ph(\d+)_pat\d+",key)
    lsize_key = "L-Size" + phase_pattern.group(0)
    phase_nr = phase_pattern.group(1)
    phase = phases[phase_nr]
    for i, gsizeitem in enumerate(gsize):
      phase_nr2 = re.search("_ph(\d+)_pat\d+",gsizeitem).group(1)
      if phase_nr2 == phase_nr:
        gsize.pop(i)
        break
    Y, sigma_Y = refined_par[key]
    Sz,sigma_Sz = refined_par[lsize_key] if lsize_key in refined_par.keys() else (0,0)

    """ calculating the ab/c sizes. """
    Y_degree = Y
    Y_AA = Y_degree * math.pi**2 / 180 / 2 / wavelength * 1000
    sigma_Y = sigma_Y * math.pi ** 2 / 180 / 2 / wavelength * 1000

    try:
      absize = 100 / Y_AA
      sigma_absize = 100 * (sigma_Y ** 2 / Y_AA ** 4) ** 0.5
      csize = 100 / (Y_AA + Sz)
      sigma_csize = 100 * ((sigma_Y ** 2 + sigma_Sz ** 2) / (Y_AA + Sz) ** 4) ** 0.5
    except ZeroDivisionError:
      absize = np.nan
      csize = np.nan
      sigma_absize = np.nan
      sigma_csize = np.nan
    """ prepare dictionary """
    tmp_dict = {"app.cry. ab-size": (absize, sigma_absize), "app.cry. c-size": (csize, sigma_csize)}
    res_dict[phase] = tmp_dict

  """ the calculation performed below are done according to the docstring explanation """
  for key in gsize:
    """ find calculated appearent size from .mic """
    mics = findfiles(".mic")
    phase_pattern = re.search("_ph(\d+)_pat\d+",key)
    phase_nr = phase_pattern.group(1)
    phase = phases[phase_nr]
    mics = [mic for mic in mics if mic[-5] == phase_nr]
    for mic in mics:
      with open(mic, "r") as reading:
        for line in reading:
          if "Phase No:" in line:
            phasename = re.search("Phase No:\s*\d+\s*(.*)", line)
            phasename = phasename.group(1).strip()
          if "standard deviation (anisotropy):" in line:
            found = re.search(":\s*([^\s]+)\s*\(\s*([^\s]+)\)", line)
            size,error = found.group(1),found.group(2)
            absize,sigma_absize = (float(size)/10.0, float(error)/10.0) # div 10 to go Å to nm.
            csize,sigma_csize = (float(size)/10.0, float(error)/10.0) # div 10 to go Å to nm.
            break
    tmp_dict = {"app.cry. ab-size": (absize, sigma_absize), "mic": (csize, sigma_csize)}
    res_dict[phase] = tmp_dict

  return res_dict

def find_significant_figure(error_value_as_string):
  """
  Calculates the number of signification digits allowed from a given error.

  Args:
      error_value_as_string: String of an error value

  Returns:
      length of error value given as edible integer index for the decimal.Decimal module
      Is used to shorten the actual value.

  Examples:
      Eg. value = 1.0034 but error = 0.1, then returned index will be
      1. Value is displayed as 1.0(1).

      Eg. value = 0.1001 and error = 2.0, then returned index is -1.
      Value is displayed as 0(2).

  """
  try:
    error_value_as_string = str(error_value_as_string)
    if error_value_as_string[0] == "-":
      error_value_as_string = error_value_as_string[1:]
    if not re.search("\.", error_value_as_string):
      return -len(error_value_as_string)+1
    if error_value_as_string[0] != "0":
      return -len(error_value_as_string.split(".")[0])+1
    error_value_as_string = error_value_as_string.rstrip("0")
    no_period = re.sub("\.","",error_value_as_string)
    integer = int(no_period)
    return len(no_period) - len(str(integer))
  except Exception as e:
    import pdb; pdb.set_trace(); print(e); print("error_value_as_string =", error_value_as_string, "\n")

def round2significant_error(value, error, length = ""):
  """ Combines a value and its error to a compactly formatted string.

  Args:
    value: the refined parameter value
    error: the error of the refined parameter value.
    length: the maximum allowed lenght of return string

  Returns:
    String that has been formatted to only to the point of
    the significant figure according to the FullProf provided error.
    See examples below.

  example 1:
    value = 34.5623
    error = 0.03
    length = 15
    output = 34.56(3)

  example 2:
    value = -34.5623
    error = 11.03
    length = 15
    output = -3(1)e+01 --> always at least 9 chars to describe equal length value/error.

  example 3:
    value = 34.5623
    error = 0.03
    length = 5
    output = 35(0)

    """
  if error is not np.nan:
    error = abs(error)
  strvalue = "{:.100f}".format(value)
  strerror = "{:.100f}".format(error)
  if (value == 0.0) and (error == 0.0):
    return "0(0)"
  if (value is np.nan) or (error is np.nan):
    return "NaN"
  if (str(value).lower() == "nan") or (str(error).lower() == "nan"):
    return "NaN"
  if length != str():
    roundvalue = round(Decimal(strvalue), length)
    rounderror = round(Decimal(strerror), length)
    result = round2significant_error(float(roundvalue), float(rounderror))
    i = 1
    while len(result) > length:
      strvalue = "{:.100f}".format(float(roundvalue))
      strerror = "{:.100f}".format(float(rounderror))
      roundvalue = round(Decimal(strvalue), length-i)
      rounderror = round(Decimal(strerror), length-i)
      result = round2significant_error(float(roundvalue), float(rounderror))
      i+=1
    return result
  if error == 0:
    if "e" in str(value):
      """ this option called if round2significant_error calls itself
      with a zero error. """
      pre,post = str(value).split("e")
      return "{}{}e{}".format(pre,"(0)",post)
    else:
      return "{}({})".format(str(value),"0")

  sign = "" if value >= 0 else "-"
  strvalue = "{:.100f}".format(value)
  strerror = "{:.100f}".format(error)
  significant_value = find_significant_figure(strvalue)
  significant_error = find_significant_figure(strerror)
  roundvalue = round(Decimal(strvalue), significant_error)
  rounderror = round(Decimal(strerror), significant_error)
  if significant_error < 0:
    rounderror = "{:.100f}".format(float(rounderror))
    significant_error = find_significant_figure(rounderror)
    significant_value = find_significant_figure(strvalue)
    roundvalue = round(Decimal(strvalue), significant_error)
    rounderror = round(Decimal(strerror), significant_error)
    exponent = str(-significant_error) if abs(significant_error) > 10 else "0" + str(-significant_error)
    index = len(str(abs(int(roundvalue)))) - len(str(int(rounderror))) + 1
    if index == 0:
      index = 1
      roundvalue = 0
    string = "{}{}({})e+{}".format(sign,str(int(abs(roundvalue)))[:index],str(rounderror)[0],exponent)
    return string
  else:
    if abs(roundvalue) <= abs(rounderror):
      roundvalue = round(Decimal(strvalue), significant_error)
      rounderror = round(Decimal(strerror), significant_error)
      if roundvalue != 0:
        roundvalue = str(roundvalue).rstrip(".0")
      if rounderror != 0:
        rounderror = str(rounderror).rstrip(".0")
    """ investigate if new significant figure of error,
    if so, update roundvalue and rounderror """
    if "E" in str(rounderror):
      newposition = int(re.search("E[+-0](\d+)",str(rounderror)).group(1))
      rounderror = re.search("(\d)[.\d]*E",str(rounderror)).group(1)
    else:
      # try:
      #     rounderror = "{:.20f}".format(float(rounderror))
      # except:
      #     import pdb; pdb.set_trace(); print("rounderror =",rounderror)
      newposition = find_significant_figure(rounderror)
      rounderror = str(round(Decimal(rounderror),newposition))
      if "." in rounderror:
        rounderror = rounderror[1+newposition]
      if "E" in rounderror:
        """ ugly workaround """
        pre,post = rounderror.split("E")
        tmp_post = int(post)
        roundvalue = int((float(roundvalue)-(float(roundvalue)%10**tmp_post))/10**tmp_post)
        if len(str(tmp_post)) == 1:
          post = post[0] + "0" + post[1]
        return "{}({})e{}".format(roundvalue,pre,post)
    roundvalue = round(Decimal(roundvalue),newposition)
    """ taking care that roundvalue is not expressed in scientific terms... """
    if "E" in str(roundvalue):
      roundvalue = "{:.20f}".format(float(roundvalue)).rstrip("0")
    return "{}({})".format(roundvalue,rounderror)

@TraceCalls()
def prepare_refinement_result(phases = "", refined_par = "", size_info = "", frac_info = "", spacegroups = "", wavelength = ""):
  """
  preparing information such as refined parameters, phases, phase fractions
  crystallite sizes and so to be printed on canvas.

  Args:
    phases: Dictionary
      keys: integer number
      values: phase names

    refined_par: Dictionary
      keys: FullProf named parameter
      values: tuple of value and its error

    size_info: dictionary
      keys: phase name
      values:
        tuple of ab and c appearent crystallite sizes
        calculated from Lorentzian line broadening and (if available)
        Lorenzian anisotropic parameter

    frac_info: dictionary of phase weight fractions.
      keys: phase name.
      values: value and its error.

    spacegroups: dictionary
      keys: integer number
      values: string of spacegroup name as written in .pcr file.

  Returns:
    string formatted to be printed on canvas, non-phase parameters on the top,
    then each phase (listed in sorted order) is listed with corresponding parameters.

    No support for showing linked parameters.
    Parameters associated with more phases are only plotted once.

  """

  """ preparing entry in dictionary for non-phase related pars """
  insert_text = {"0": []}
  """ preparing entries in a dictionary for information on the different phases """
  for phase_key in phases.keys():
    insert_text[phase_key] = ["{:16.15} {}".format(spacegroups[phase_key], phases[phase_key])]

  """ subfunction to prepare text for insertion """
  def insert_text2(newline, name, value, error):
    """
    Prepares a single stringline. Values and errors are rounded up.

    Args:
      newline: '\n' if needed.

      name: commonly the name of refined parameter. can also be phase name etc

      value: the predicted value from FP

      error: the calculated error from FP. The first non-zero index from error is used to determine the maximum shown value of 'value'.
    Returns:
      a single stringline that has name of refined parameter and value + error.

      e.g. value = 8.357975 error = 0.00049791916 for refined parameter Cell_A_ph1_pat is shown as

      'Cell_A_ph1_pat  8.3580(5)   [$\\AA$]\n'

      The (5) should be translated to 8.3580 +/- 0.0005

    """

    """ note on .format
    https://pyformat.info/

    note on strftime
    http://strftime.org/
    """

    """ finding correct unit for parameter """
    unitdict = {"unit" : "[$\\AA$]", "weight fraction" : "[%]", "app." : "[nm]"}
    unit = ""
    for key in unitdict.keys():
      if name.startswith(key):
        unit = unitdict[key]
        break
    """ checking if R wp / chi2 """
    if (name == "R wp / chi2"):
      """ restructuring the R wp and chi2 manually... """
      return "{}{}{:<11.10}{}{}{}".format(r"R$_{wp}$"," = ", value, r"$\chi^2$"," = ", error)
      # return "{}{:17.16}{:3.3} / {:3.3}".format(newline, name, value, error )
    string = round2significant_error(value, error, 13) # 13 corresponce to showing eight characters of str(value).
    result = "{}{:17.16}{:14.13}{}".format(newline, name, string, unit)
    return result
    # END of internal function

  """ unpacking refined parameters phase-wise, skipping background pars """
  for key, (value, error) in refined_par.items():
    if "bck_" in key.lower():
      continue
    phase_pattern = re.search("_ph(\d+)",key)
    if not phase_pattern:
      if key == "R wp / chi2":
        insert_text["0"].append( insert_text2("",key, value, error) )
        """ manually inserting the wavelength information """
        unit = "[$\\AA$]"
        insert_text["0"].append( "{}{:<17.16}{:<14.9}{}".format("\n","wavelength", wavelength, unit) )
      else:
        key = key.split("_pat")[0]
        if key == "Zero":
          key = "zero shift"
        if key == "zero shift":
          """ I decided to stop showing the zero shift. adds no information... """
          continue
        insert_text["0"].append( insert_text2("\n",key, value, error) )
      continue
    """ if parameter not special, we transform to appropiate names """
    key = key.split("_ph")[0]

    key = "debye-waller f." if key == "Bover" else key
    key = "scale factor" if key == "Scale" else key
    key = "lor. line broad." if key == "Y-cos" else key
    key = "unit cell A" if key == "Cell_A" else key
    key = "unit cell C" if key == "Cell_C" else key
    key = "unit cell B" if key == "Cell_B" else key
    key = "lor. size distr." if key == "L-Size" else key

    """ inserting refined parameters from FullProf """
    insert_text[phase_pattern.group(1)] += [insert_text2("\n",key, value, error)]
  """ inserting phase fraction informations """
  reversed_phases = {y:x for x,y in phases.items()}
  for key, (value,error) in frac_info.items():
    insert_text[reversed_phases[key]] += [insert_text2("\n","weight fraction", value, error)]
  """ inserting crystallite size information """
  for key, sizes in size_info.items():
    if "mic" in sizes.keys():
      value, error = sizes["app.cry. ab-size"]
      insert_text[reversed_phases[key]] += [insert_text2("\n","mic-abc-size", value, error)]
    elif sizes["app.cry. ab-size"] == sizes["app.cry. c-size"]:
      value, error = sizes["app.cry. ab-size"]
      insert_text[reversed_phases[key]] += [insert_text2("\n","app. cryst. size", value, error)]
    else:
      for size in sizes:
        value, error = size_info[key][size]
        insert_text[reversed_phases[key]] += [insert_text2("\n",size, value, error)]
  """ combine all the different phase informations """
  combined_text = ""
  """ sorting phases for consistent output """
  tmp_keys = [key for key in insert_text.keys()]
  tmp_keys.sort(key = lambda x: insert_text[x][0].split("   ")[-1].split())
  tmp_keys.remove("0")
  tmp_keys = ["0"] + tmp_keys

  for key in tmp_keys:
    combined_text += insert_text[key].pop(0)
    tmp_list = [value for value in insert_text[key]]
    """ sorting the output according primarily to the first character
    and secondarily to the last character (if unit is included, parameter is put last)"""
    tmp_list.sort(key = lambda x: x[1::len(x)-2][::-1])
    combined_text += "".join(tmp_list) + "\n\n"

  return combined_text

# END



# bragg_peak_indications.py

from math import cos, sqrt, sin, asin, pi
import numpy as np, os, sys
sys.path.append( r"C:\AU-PHD\General_Data\_python\PyPi\magnetmatter\magnetmatter\modules" )

from extract_pcr import getwavelength
from find_tools import findfile_abspath

""" things to remember:
forbindden reflections for e.g. BCC, FCC, diamond, triangular lattice

"""

""" where did I get the formulars for the lattice spacing d?
wikipedia, yes: https://en.wikipedia.org/wiki/Crystal_structure
"""

class Crystals():
  """ to contain all Crystal objects. """
  def __init__(self, abs_path_to_pcr = ""):
    self.crystals = list()
    if abs_path_to_pcr == "":
      self.pcr = findfile_abspath(".pcr")
    else:
      self.pcr = abs_path_to_pcr
    self.wavelength = getwavelength(self.pcr)
    self.parse_pcr()

  def parse_pcr(self):
    crystals = list()
    with open(self.pcr, "r") as reading:
      for line in reading:
        if "!  Data for PHASE number: " in line:
          reading.__next__() # not important line.
          phasename = reading.__next__().strip()
        if "!     a          b         c " in line:
          """
          !     a          b         c        alpha      beta       gamma      #Cell Info
             5.144739   5.144739  14.266463  90.000000  90.000000 120.000000
          """
          line = reading.__next__()
          unitcell_dimensions = tuple(line.split()[0:6])
          mycrystal = Crystal( *(phasename, self.wavelength, *unitcell_dimensions) )
          # will this set all crystals to same object?
          # no, it does not appear to do so
          self.crystals.append( mycrystal )



class Crystal():
  """ keep track of crystal system
  optimally we search .pcr file for needed information
  when this object is initiated...

  how to handle multiple phases?
    shall these be contained within the same Crystal object?
  how to handle magnetic phases?
    ignore
  """
  def __init__(self, phasename, wavelength, a,b,c,alpha,beta,gamma):
    """ is it possible to give this information as a tuple? """
    self.phasename = phasename
    self.wavelength = wavelength
    self.a = float(a)
    self.b = float(b)
    self.c = float(c)
    self.alpha = float(alpha)
    self.beta = float(beta)
    self.gamma = float(gamma)
    self.nb_hkl_values = 999
    self.d = np.empty((self.nb_hkl_values,4)) # keys = hkl (int), value = d (float)
    """ two below lines could be moved to calculate_d """
    calculate_d = self.get_d_function()
    for hkl in range(0,self.nb_hkl_values):
      strhkl = "00" + str(hkl+1)
      strhkl = strhkl[-3:]
      self.d[hkl,0:3] = tuple( int(i) for i in list(strhkl) )
      # print("hkl =",hkl)
      # print("self.d[hkl,0:3] =",self.d[hkl,0:3])
      # print("calculate_d( *tuple(self.d[hkl,0:3]) ) =",calculate_d( *tuple(self.d[hkl,0:3]) ))
      # import pdb; pdb.set_trace();
      self.d[hkl,3] = calculate_d( *tuple(self.d[hkl,0:3]) ) # distances according to crystal system. could be optimized by directly inputting values.
      # import pdb; pdb.set_trace();
      self.d[hkl,3] = self.calculate_two_theta(self.d[hkl,3])

    """ dropping all 2theta entries of NaN """
    self.d = self.d[~np.isnan(self.d)[:,3]]
    """ sorting """
    self.d = self.d[self.d[:,3].argsort()]
    """ unique reflections below 180 degree """
    self.unique = np.unique(self.d[:,3])


    # self.d = np.flipud(self.d)
    # self.d[:,0] = range(0,1000) # all hkl values. needed to remove forbidden reflections & duplicates.

  def get_d_function(self):
    """
    inputs:
      lattice parameter a, b, c
      angles alpha, beta, gamma
    """
    if (self.a == self.b and self.b == self.c and self.alpha == 90, self.beta == 90 and self.gamma == 90):
      return self.cubic_d
    elif (self.a == self.b and self.b != self.c and self.alpha == 90 and self.beta == 90 and self.gamma == 90):
      return self.tetragonal_d
    elif (self.a == self.b and self.alpha == 90 and self.beta == 90 and self.gamma == 120):
      return self.hexagonal_d

  def calculate_two_theta(self, d):
    """ based on Bragg's law
    2d sin theta = n lambda

    we make d a function of hkl
    (which include the higher harmonics corresponding to n != 1)

    2 d sin(theta) = lambda

    asin(lambda / 2 / d) = theta
    """
    try:
      """ something is wrong with radians vs degrees... """
      return asin(self.wavelength / 2 / d) * 360 / pi # returns 2theta.
    except:
      """ lambda / 2 / d > 1, that means 2 theta > 180 which is unphysically. """
      return np.nan





  def cubic_d(self, h,k,l):
    """ remember forbidden reflections on body-centered and face-centered.
    a = b = c
    alpha = beta = gamma = 90deg
    """
    return self.a / (sqrt( h**2 + k**2 + l**2 ))

  def tetragonal_d(self):
    inside = ( (h**2 + k**2) / self.a**2 ) + l**2/self.c**2
    d = sqrt( 1 / inside )

  def hexagonal_d(self, h,k,l):
    inside = 4/3 * (h**2 + h*k + k**2)/ self.a**2 + l**2 / self.c**2
    d = sqrt( 1 / inside )

  def rhombohedral_d(self):
    top = self.a**2 * (1 - 3* cos(self.alpha)**2 + 2 * cos(self.alpha)**3)
    bottom = (h**2 + k**2 + l**2) * sin(self.alpha)**2 + 2*(h*k+k*l+h*l)*(cos(self.alpha)**2 - cos(self.alpha))
    d = math.sqrt( top / bottom )

  def orthorhombic_d(self):
    inside = h**2/self.a**2 + k**2 / self.b**2 + l**2/self.c**2
    d = sqrt( 1 / inside)

  """ other systems include
  monoclinic
  triclinic
  """
# END

if __name__ == "__main__":
  abs_pcr_path = r"C:\AU-PHD\General_Data\Report R1D1\furnaces\Ofurnace\TF10_R1D1 - hematite\_aFe2O3_Fe3O4_.pcr"
  mycrystrals = Crystals(abs_pcr_path)
  import pdb; pdb.set_trace();



# auxiliary.py

import re

def cm2inch(*tupl):
    """
    Helper function to convert figsize-cm-tuple to inches.

    Args:
        tuple or list of tupe that will have all its elements converted to
        metric dimensions.

    Returns:
        A new tuple with metric dimensions.

    """
    inch = 2.54
    if isinstance(tupl[0], tuple):
        return tuple(i/inch for i in tupl[0])
    else:
        return tuple(i/inch for i in tupl)

def natural_keys(text):
    """
    Returns a list of chars and digits that can be used for sorting algorithms.

    Args:
        text: string containing any symbols.

    Returns:
        A list of symbols that - if possible - has it character converted to
        integer datatype to allow human-like sorting.

    Usecase 1:
        alist.sort(key=natural_keys)
    Usecase 2:
        alist = sorted(alist, key = natural_keys)

    """
    atoi = lambda c: int(c) if c.isdigit() else c
    return [ atoi(c) for c in re.split('(\d+)', text) ]
